from typing import List, Optional

from google.cloud import texttospeech_v1
from google.cloud.texttospeech_v1.types import (
    AudioConfig,
    VoiceSelectionParams,
    SynthesisInput,
)

from . import (
    LibraryContext,
    keyword,
)


class TextToSpeechKeywords(LibraryContext):
    """Class for Google Cloud Text-to-Speech API

    Link to `Text To Speech PyPI`_ page.

    .. _Text To Speech PyPI: https://pypi.org/project/google-cloud-texttospeech/
    """

    def __init__(self, ctx):
        super().__init__(ctx)
        self.service = None

    @keyword(tags=["init", "text to speech"])
    def init_text_to_speech(
        self,
        service_account: str = None,
        use_robocorp_vault: Optional[bool] = None,
        token_file: str = None,
    ) -> None:
        """Initialize Google Cloud Text to Speech client

        :param service_account: file path to service account file
        :param use_robocorp_vault: use credentials in `Robocorp Vault`
        :param token_file: file path to token file
        """
        self.service = self.init_service_with_object(
            texttospeech_v1.TextToSpeechClient,
            service_account,
            use_robocorp_vault,
            token_file,
        )

    @keyword(tags=["text to speech"])
    def list_supported_voices(self, language_code: str = None) -> List:
        """List supported voices for the speech

        :param language_code: voice languages to list, defaults to None (all)
        :return: list of supported voices

        **Examples**

        **Robot Framework**

        .. code-block:: robotframework

            ${result}=   List Supported Voices   en-US
        """
        if language_code:
            voices = self.service.list_voices(language_code)
        else:
            voices = self.service.list_voices()
        return voices.voices

    @keyword(tags=["text to speech"])
    def synthesize_speech(
        self,
        text: str,
        language: str = "en-US",
        name: str = "en-US-Standard-B",
        gender: str = "MALE",
        encoding: str = "MP3",
        target_file: str = "synthesized.mp3",
    ) -> List:
        """Synthesize speech synchronously

        :param text: input text to synthesize
        :param language: voice language, defaults to "en-US"
        :param name: voice name, defaults to "en-US-Standard-B"
        :param gender: voice gender, defaults to "MALE"
        :param encoding: result encoding type, defaults to "MP3"
        :param target_file: save synthesized output to file,
            defaults to "synthesized.mp3"
        :return: synthesized output in bytes

        **Examples**

        **Robot Framework**

        .. code-block:: robotframework

            ${result}=   Synthesize Speech   ${text}
        """
        synth_input = SynthesisInput(text=text)
        voice_selection = VoiceSelectionParams(
            language_code=language, name=name, ssml_gender=gender
        )
        audio_config = AudioConfig(audio_encoding=encoding)
        response = self.service.synthesize_speech(
            request={
                "input": synth_input,
                "voice": voice_selection,
                "audio_config": audio_config,
            }
        )
        if target_file:
            with open(target_file, "wb") as f:
                f.write(response.audio_content)
        return response.audio_content



import base64
import json
import os
import pickle
import tempfile
from pathlib import Path

from apiclient import discovery
from google_auth_oauthlib.flow import InstalledAppFlow
from google.auth.transport.requests import Request
from google.oauth2 import service_account as oauth_service_account


class ElementNotFound(ValueError):
    """No matching elements were found."""


class MultipleElementsFound(ValueError):
    """Multiple matching elements were found, but only one was expected."""


class TimeoutException(ValueError):
    """Timeout reached while waiting for condition."""


class GoogleOAuthAuthenticationError(Exception):
    """Raised when unable to get Google OAuth credentials."""


class LibraryContext:
    """Shared context for all keyword libraries."""

    def __init__(self, ctx):
        self.ctx = ctx

    @property
    def logger(self):
        return self.ctx.logger

    @property
    def robocorp_vault_name(self):
        return self.ctx.robocorp_vault_name

    @property
    def robocorp_vault_secret_key(self):
        return self.ctx.robocorp_vault_secret_key

    @property
    def use_robocorp_vault(self):
        return self.ctx.use_robocorp_vault

    @property
    def service_account_file(self):
        return self.ctx.service_account_file

    @property
    def cloud_auth_type(self):
        return self.ctx.cloud_auth_type

    def get_secret_from_robocorp_vault(self, secret_type="serviceaccount"):
        if self.ctx.secrets_library is None:
            raise KeyError("RPA.Robocorp.Vault library is required to use Vault")

        if (
            self.ctx.robocorp_vault_name is None
            or self.ctx.robocorp_vault_secret_key is None
        ):
            raise KeyError(
                "Both 'robocorp_vault_name' and 'robocorp_vault_secret_key' "
                "are required to access Robocorp Vault. Set them in library "
                "init or with `set_robocorp_vault` keyword."
            )

        vault_items = self.ctx.secrets_library().get_secret(
            self.ctx.robocorp_vault_name
        )
        secret = vault_items[self.ctx.robocorp_vault_secret_key]
        if secret_type == "serviceaccount":
            secret_obj = (
                secret if isinstance(secret, dict) else json.loads(secret.strip())
            )
            with tempfile.NamedTemporaryFile(mode="w", delete=False) as temp_filedesc:
                json.dump(secret_obj, temp_filedesc, ensure_ascii=False)
                return temp_filedesc.name
        else:
            return secret

    def init_service(
        self,
        service_name: str,
        api_version: str,
        scopes: list,
        service_account_file: str = None,
        credentials_file: str = None,
        use_robocorp_vault: bool = False,
        token_file: str = None,
        save_token: bool = False,
        auth_type: str = None,
    ) -> None:
        """Initialize Google Service

        :param service_account: filepath to credentials JSON
        :param use_robocorp_vault: use json stored into `Robocorp Vault`
        """
        service = None
        credentials = None
        self.logger.debug("Init service with scopes: %s", scopes)
        scopes = [f"https://www.googleapis.com/auth/{scope}" for scope in scopes]
        if use_robocorp_vault is not None:
            use_cloud = bool(use_robocorp_vault)
        else:
            use_cloud = self.ctx.use_robocorp_vault
        if use_cloud:
            cloud_auth_type = auth_type or self.ctx.cloud_auth_type
            credentials = self.get_credentials_from_robocorp_vault(
                cloud_auth_type,
                service_account_file,
                scopes,
                token_file,
                credentials_file,
                save_token,
            )
        elif service_account_file:
            self.logger.info("Authenticating with service account file")
            credentials = oauth_service_account.Credentials.from_service_account_file(
                service_account_file, scopes=scopes
            )
        elif token_file:
            self.logger.info("Authenticating with oauth token file")
            credentials = self.get_credentials_with_oauth_token(
                use_cloud,
                token_file,
                credentials_file,
                scopes,
                save_token,
            )
        elif self.ctx.service_account_file:
            self.logger.info("Authenticating with service account file")
            credentials = oauth_service_account.Credentials.from_service_account_file(
                self.ctx.service_account_file, scopes=scopes
            )

        try:
            service = discovery.build(
                service_name,
                api_version,
                credentials=credentials,
                cache_discovery=False,
            )
        except OSError as e:
            raise AssertionError from e

        if service is None:
            raise AssertionError("Failed to create service")
        return service

    def init_service_with_object(
        self,
        client_object: str,
        service_account_file: str,
        use_robocorp_vault: bool,
        token_file: str = None,
        auth_type: str = None,
        **kwargs,
    ):
        service = None
        if use_robocorp_vault is not None:
            robocorp_vault = bool(use_robocorp_vault)
        else:
            robocorp_vault = self.ctx.use_robocorp_vault

        cloud_auth_type = auth_type or self.ctx.cloud_auth_type
        if robocorp_vault:
            service = self.get_service_from_robocorp_vault(
                client_object, cloud_auth_type, service_account_file, **kwargs
            )
        elif service_account_file:
            self.logger.info("Authenticating with service account file")
            service = client_object.from_service_account_json(
                service_account_file, **kwargs
            )
        elif token_file:
            self.logger.info("Authenticating with oauth token file")
            token_file_location = Path(token_file).absolute()
            if os.path.exists(token_file_location):
                with open(token_file_location, "rb") as token:
                    credentials = pickle.loads(token)
                    service = client_object(credentials=credentials, **kwargs)
        elif self.ctx.service_account_file:
            self.logger.info("Authenticating with service account file")
            service = client_object.from_service_account_json(
                self.ctx.service_account_file, **kwargs
            )
        else:
            self.logger.info("Authenticating with default client object")
            service = client_object(**kwargs)

        if service is None:
            raise AssertionError("Failed to create service")

        return service

    def write_json(self, json_file, response):
        if json_file and response:
            with open(json_file, "w") as f:  # pylint: disable=unspecified-encoding
                f.write(response.__class__.to_json(response))

    def get_credentials_with_oauth_token(
        self, use_robocorp_vault, token_file, credentials_file, scopes, save_token
    ):
        credentials = None
        if use_robocorp_vault:
            token = self.get_secret_from_robocorp_vault("token")
            credentials = pickle.loads(base64.b64decode(token))
        else:
            token_file_location = Path(token_file).absolute()
            if os.path.exists(token_file_location):
                with open(  # pylint: disable=unspecified-encoding
                    token_file_location, "r"
                ) as token:
                    credentials = pickle.loads(base64.b64decode(token.read()))
        if not credentials or not credentials.valid:
            if credentials and credentials.expired and credentials.refresh_token:
                credentials.refresh(Request())
            else:
                flow = InstalledAppFlow.from_client_secrets_file(
                    credentials_file, scopes
                )
                credentials = flow.run_local_server()
            if save_token:
                with open(token_file_location, "wb") as token:
                    pickle.dump(credentials, token)
        if not credentials:
            raise GoogleOAuthAuthenticationError(
                "Could not get Google OAuth credentials"
            )
        return credentials

    def get_credentials_from_robocorp_vault(
        self,
        cloud_auth_type,
        service_account_file,
        scopes,
        token_file,
        credentials_file,
        save_token,
    ):
        credentials = None
        if cloud_auth_type == "serviceaccount":
            self.logger.info(
                "Authenticating with service account file from Robocorp Vault"
            )
            service_account_file = self.get_secret_from_robocorp_vault("serviceaccount")
            credentials = oauth_service_account.Credentials.from_service_account_file(
                service_account_file, scopes=scopes
            )
        else:
            self.logger.info("Authenticating with oauth token file from Robocorp Vault")
            credentials = self.get_credentials_with_oauth_token(
                True,
                token_file,
                credentials_file,
                scopes,
                save_token,
            )
        return credentials

    def get_service_from_robocorp_vault(
        self, client_object, cloud_auth_type, service_account_file, **kwargs
    ):
        service = None
        if cloud_auth_type == "serviceaccount":
            try:
                self.logger.info(
                    "Authenticating with service account file from Robocorp Vault"
                )
                service_account_file = self.get_secret_from_robocorp_vault(
                    "serviceaccount"
                )
                service = client_object.from_service_account_json(
                    service_account_file, **kwargs
                )
            finally:
                if service_account_file:
                    os.remove(service_account_file)
        else:
            self.logger.info("Authenticating with oauth token file from Robocorp Vault")
            token = self.get_secret_from_robocorp_vault("token")
            credentials = pickle.loads(base64.b64decode(token))
            service = client_object(credentials=credentials, **kwargs)
        return service



import mimetypes
import pickle
from typing import List, Optional

from google.api_core.client_options import ClientOptions
from google.cloud import documentai_v1 as documentai

from . import LibraryContext, keyword


class DocumentAIKeywords(LibraryContext):
    """Keywords for Google Cloud Document AI service.

    Google Document AI provides processors for intelligent
    document processing (IDP).

    To take Document AI into use:

        - Create Google Cloud project
        - Enable Google Cloud Document AI API for the project
        - For Document AI product page, activate desired processors
          for the project
    """

    def __init__(self, ctx):
        super().__init__(ctx)
        self.service = None

    @keyword(tags=["init", "document ai"])
    def init_document_ai(
        self,
        service_account: Optional[str] = None,
        region: Optional[str] = "us",
        use_robocorp_vault: Optional[bool] = None,
        token_file: Optional[str] = None,
    ) -> None:
        """Initialize Google Cloud Document AI client

        :param service_account: file path to service account file
        :param region: region of the service
        :param use_robocorp_vault: use credentials in `Robocorp Vault`
        :param token_file: file path to token file

        Example:

        Robot Framework example:

            # Init using Service Account from a file
            Init Document AI   ${CURDIR}${/}service_account.json  region=eu
            # Init using OAuth token from a file and default "us" region
            Init Document AI   ${CURDIR}${/}token.json
            # Init using service account file from the Robocorp Vault
            Set Robocorp Vault
            ...         vault_name=DocumentAI
            ...         vault_secret_key=google-sa
            Init Document AI    region=eu    use_robocorp_vault=True

        Python example:

        .. code-block:: python

            GOOGLE = Google()
            GOOGLE.set_robocorp_vault("DocumentAI", "google-sa")
            GOOGLE.init_document_ai(region="eu", use_robocorp_vault=True)
        """
        kwargs = {}
        if not region:
            raise ValueError("Parameter 'region' needs to point to a service region")
        if region.lower() != "us":
            opts = ClientOptions(
                api_endpoint=f"{region.lower()}-documentai.googleapis.com"
            )
            kwargs["client_options"] = opts
        self.logger.info(f"Using Document AI from '{region.upper()}' region")
        self.service = self.init_service_with_object(
            documentai.DocumentProcessorServiceClient,
            service_account,
            use_robocorp_vault,
            token_file,
            **kwargs,
        )

    @keyword(tags=["document ai"])
    def process_document(
        self,
        project_id: str,
        region: str,
        processor_id: str,
        file_path: str,
        mime_type: Optional[str] = None,
    ) -> documentai.Document:
        """Process document in the Google Cloud platform
        using given document processor ID within given project and
        region.

        For a full list of Document response object attributes, please reference this
        `page <https://cloud.google.com/python/docs/reference/documentai/latest/google.cloud.documentai_v1.types.Document/>`_.


        :param project_id: Google Cloud project ID
        :param region: Google Cloud region of the processor
        :param processor_id: ID of the document processor
        :param file_path: filepath of the document to process
        :param mime_type: given mime type of document (optional),
         if not given it is auto detected
        :return: processed document response object

        Example:

        Robot Framework example:

            ${document}=    Process Document
            ...    project_id=${GOOGLE_PROJECT_ID}
            ...    region=eu
            ...    processor_id=${RECEIPT_PROCESSOR_ID}
            ...    file_path=${CURDIR}${/}mydocument.pdf
            ${entities}=    Get Document Entities    ${document}
            FOR  ${ent}  IN  @{entities}
                Log To Console  Entity: ${ent}
            END
            ${languages}=    Get Document Languages    ${document}
            Log To Console    Languages: ${languages}

        Python example:

        .. code-block:: python

            document = GOOGLE.process_document(
                project_id=PROJECT_ID,
                region="eu",
                processor_id=PROCESSOR_ID,
                file_path="./files/mydocument.pdf",
            )
            entities = GOOGLE.get_document_entities(document)
            for ent in entities:
                print(ent)
            languages = GOOGLE.get_document_languages(document)
            for lang in languages:
                print(lang)
        """  # noqa: E501
        name = self.service.processor_path(project_id, region, processor_id)

        # Read the file into memory
        with open(file_path, "rb") as binary:
            binary_content = binary.read()

        mime = mime_type or mimetypes.guess_type(file_path)[0]
        self.logger.info(f"Processing document '{file_path}' with mimetype '{mime}'")
        # Load Binary Data into Document AI RawDocument Object
        raw_document = documentai.RawDocument(content=binary_content, mime_type=mime)

        # Configure the process request
        request = documentai.ProcessRequest(name=name, raw_document=raw_document)

        result = self.service.process_document(request=request)

        document = result.document
        return document

    @keyword(tags=["document ai"])
    def save_document_response(
        self, document: documentai.Document, filepath: str
    ) -> None:
        """Save ``Process Document`` response into a binary file.

        :param document: response document object
        :param filepath: target file to save binary object into

        Robot Framework example:

        .. code-block:: robotframework

            ${document}=    Process Document
            ...    project_id=101134120147
            ...    region=eu
            ...    processor_id=${RECEIPT_PROCESSOR}
            ...    file_path=${file_in}
            # save response for later
            Save Document Response  ${CURDIR}${/}google_processed.response

        Python example:

        .. code-block:: python

            document = GOOGLE.process_document(
                project_id=PROJECT_ID,
                region="eu",
                processor_id=PROCESSOR_ID,
                file_path="./files/receipt1.jpg",
            )
            GOOGLE.save_document_response(document, "receipt.response")
        """
        with open(filepath, "wb") as response_file:
            pickle.dump(document, response_file)

    @keyword(tags=["document ai"])
    def load_document_response(self, filepath: str) -> documentai.Document:
        """Loads the binary object saved by ``Save Document Response`` into
        ``documentai.Document`` format which is accessible by helper keywords.

        :param filepath: source file to read binary document object from
        :return: processed document response object

        Robot Framework example:

        .. code-block:: robotframework

            # load previously saved response
            ${document}=  Load Document Response  ${CURDIR}${/}google_processed.response
            ${entities}=  Get Document Entities  ${document}

        Python example:

        .. code-block:: python

            document = GOOGLE.load_document_response("google_doc.response")
            entities = GOOGLE.get_document_entities(document)
            for ent in entities:
                print(ent)
        """
        document = None
        with open(filepath, "rb") as response_file:
            try:
                document = pickle.load(response_file)
            except pickle.UnpicklingError as err:
                raise ValueError from err(
                    "The file '%s' is not 'documentai.Document' type" % filepath
                )

        if not isinstance(document, documentai.Document):
            raise ValueError(
                "The file '%s' is not 'documentai.Document' type" % filepath
            )
        return document

    @keyword(tags=["document ai", "get"])
    def get_document_entities(self, document: documentai.Document) -> List:
        """Helper keyword for getting document `entities` from a ``Process Document``
        response object.

        For examples. see ``Process Document`` keyword

        :param document: the document response object
        :return: detected entities in the document response as a list
        """
        entities = []
        for ent in document.entities:
            entities.append(
                {
                    "id": ent.id,
                    "confidence": ent.confidence,
                    "page": int(ent.page_anchor.page_refs[0].page) + 1,
                    "type": ent.type_,
                    "normalized": ent.normalized_value.text,
                    "text": ent.text_anchor.content,
                }
            )
        return entities

    @keyword(tags=["document ai", "get"])
    def get_document_languages(self, document: documentai.Document) -> List:
        """Helper keyword for getting detected `languages` from a ``Process Document``
        response object.

        For examples. see ``Process Document`` keyword

        :param document: the document response object
        :return: detected languages in the document response as a list
        """
        languages = []
        for page in document.pages:
            for lang in page.detected_languages:
                languages.append(
                    {
                        "page": page.page_number,
                        "code": lang.language_code,
                        "confidence": lang.confidence,
                    }
                )
        return languages

    @keyword(tags=["document ai"])
    def list_processors(self, project_id: str, region: str):
        """List existing document AI processors from given project and region.

        Requires `documentai.processors.list` permission.

        :param project_id: Google Cloud project ID
        :param region: Google Cloud region of the processor
        :return: list of available processors as a list

        Robot Framework example:

        .. code-block:: robotframework

            @{processors}=    List Processors    ${PROJECT_ID}    eu
            FOR    ${p}    IN    @{processors}
                # name: projects/PROJECT_ID/locations/REGION/processors/PROCESSOR_ID
                Log To Console    Processor name: ${p.name}
                Log To Console    Processor type: ${p.type_}
                Log To Console    Processor display name: ${p.display_name}
            END

        Python example:

        .. code-block:: python

            processors = GOOGLE.list_processors(PROJECT_ID, "eu")
            for p in processors:
                print(f"Processor name: {p.name}")
                print(f"Processor type: {p.type_}")
                print(f"Processor name: {p.display_name}")
        """
        parent_value = self.service.common_location_path(project_id, region)
        # Initialize request argument(s)
        request = documentai.ListProcessorsRequest(
            parent=parent_value,
        )
        processor_list = self.service.list_processors(request=request)
        return processor_list



from io import BytesIO
import mimetypes
from pathlib import Path
import shutil
import time
from typing import Dict, List, Optional

from apiclient.errors import HttpError
from apiclient.http import MediaFileUpload, MediaIoBaseDownload

from . import LibraryContext, keyword, UpdateAction
from .enums import DriveRole, DriveType, to_drive_role, to_drive_type


class GoogleDriveError(Exception):
    """Raised with errors in Drive API"""


class DriveKeywords(LibraryContext):
    """Class for Google Drive API

    For more information about Google Drive API link_.

    .. _link: https://developers.google.com/drive/api
    """

    def __init__(self, ctx):
        super().__init__(ctx)
        self.service = None

    @keyword(tags=["init", "drive"])
    def init_drive(
        self,
        service_account: str = None,
        credentials: str = None,
        use_robocorp_vault: Optional[bool] = None,
        scopes: list = None,
        token_file: str = None,
    ) -> None:
        """Initialize Google Drive client

        :param service_account: file path to service account file
        :param credentials: file path to credentials file
        :param use_robocorp_vault: use credentials in `Robocorp Vault`
        :param scopes: list of extra authentication scopes
        :param token_file: file path to token file
        """
        drive_scopes = [
            "drive",
            "drive.appdata",
            "drive.file",
            "drive.install",
            "drive.metadata",
        ]
        if scopes:
            drive_scopes += scopes
        self.service = self.init_service(
            service_name="drive",
            api_version="v3",
            scopes=drive_scopes,
            service_account_file=service_account,
            credentials_file=credentials,
            use_robocorp_vault=use_robocorp_vault,
            token_file=token_file,
        )

    @keyword(tags=["drive"])
    def upload_drive_file(
        self,
        filename: str = None,
        folder: str = None,
        overwrite: bool = False,
        make_dir: bool = False,
    ) -> str:
        """Upload files into Drive

        :param filename: name of the file to upload
        :param folder: target folder for upload
        :param overwrite: set to `True` if already existing file should be overwritten
        :param make_dir: set to `True` if folder should be created if it does not exist
        :return: uploaded file id

        Example:

        .. code-block:: robotframework

            ${file1_id}=  Upload Drive File  data.json  # Upload file to drive root
            ${file2_id}=  Upload Drive File  newdata.json  new_folder  make_dir=True
            ${file3_id}=  Upload Drive File  data.json  overwrite=True
        """
        folder_id = self.get_drive_folder_id(folder)
        if folder_id is None and make_dir:
            folder = self.create_drive_directory(folder)
            folder_id = folder["id"]
        if folder_id is None:
            raise GoogleDriveError(
                "Target folder '%s' does not exist or could not be created" % folder
            )

        filepath = Path(filename)
        if filepath.is_dir():
            raise GoogleDriveError(
                "The '%s' is a directory and can not be uploaded" % filename
            )
        elif not filepath.is_file():
            raise GoogleDriveError("Filename '%s' does not exist" % filename)

        query_string = f"name = '{filepath.name}' and '{folder_id}' in parents"
        self.logger.debug("Upload query_string: '%s'" % query_string)
        target_file = self.search_drive_files(query=query_string, recurse=True)
        guess_mimetype = mimetypes.guess_type(str(filepath.absolute()))
        file_mimetype = guess_mimetype[0] if guess_mimetype else "*/*"
        media = MediaFileUpload(
            filepath.absolute(), mimetype=file_mimetype, resumable=True
        )
        file_metadata = {
            "name": filepath.name,
            "parents": [folder_id],
            "mimeType": file_mimetype,
        }
        self.logger.debug("Upload file_metadata: '%s'" % file_metadata)
        if len(target_file) == 1 and overwrite:
            self.logger.info("Overwriting file '%s' with new content", filename)
            return self._file_update(target_file, media)
        elif len(target_file) == 1 and not overwrite:
            self.logger.warn("Not uploading new copy of file '%s'", filepath.name)
            return target_file[0]["id"]
        elif len(target_file) > 1:
            self.logger.warn(
                "Drive already contains '%s' copies of file '%s'. Not uploading again."
                % (len(target_file), filepath.name)
            )
            return None
        else:
            return self._file_create(file_metadata, media)

    def _file_create(self, file_metadata, media):
        try:
            result = (
                self.service.files()
                .create(
                    body=file_metadata,
                    media_body=media,
                    fields="id",
                )
                .execute()
            )
            return result.get("id", None)
        except HttpError as err:
            raise GoogleDriveError(str(err)) from err

    def _file_update(self, target_file, media):
        try:
            result = (
                self.service.files()
                .update(fileId=target_file[0]["id"], media_body=media, fields="id")
                .execute()
            )
            return result.get("id", None)
        except HttpError as err:
            raise GoogleDriveError(str(err)) from err

    def _download_with_fileobject(self, file_object):
        try:
            request = self.service.files().get_media(fileId=file_object["id"])
        except HttpError as err:
            raise GoogleDriveError(str(err)) from err
        fh = BytesIO()
        downloader = MediaIoBaseDownload(fh, request)
        done = False
        while done is False:
            _, done = downloader.next_chunk()
        fh.seek(0)
        with open(file_object["name"], "wb") as f:
            # pylint: disable=protected-access
            shutil.copyfileobj(fh, f, length=downloader._total_size)

    @keyword(tags=["drive"])
    def download_drive_files(
        self,
        file_dict: dict = None,
        query: str = None,
        source: str = None,
        limit: int = None,
        timeout: float = None,
    ) -> List:
        """Download files specified by file dictionary or query string

        Parameters `start`, `limit` and `timeout` are used only when
        downloading files defined by `query` parameter.

        :param file_dict: file dictionary returned by `Search Drive Files`
        :param query: drive query string to find target files, defaults to None
        :param source: source directory where query is executed
        :param limit: maximum amount of files that are downloaded, defaults to None
        :param timeout: maximum allowed time in seconds for download process
        :return: list of downloaded files

        Example:

        .. code-block:: robotframework

            ${files}=    Search Drive Files    query=name contains '.json'
            FOR    ${f}    IN    @{files}
                IF  ${f}[size] < 2000
                    Download Drive Files  file_dict=${f}
                END
            END

            ${folder_id}=   Get Drive Folder Id   datafolder
            Download Drive Files  query=name contains '.json' and '${folder_id}' in parents
        """  # noqa: E501
        if query:
            filelist = self.search_drive_files(query, source=source)
            files_downloaded = []
            start_time = time.time()

            for f in filelist:
                self._download_with_fileobject(f)
                current_time = time.time()
                files_downloaded.append(f["name"])
                if limit and len(files_downloaded) >= limit:
                    self.logger.info(
                        "Drive download limit %s reached. Stopping the download.", limit
                    )
                    break
                if timeout and (current_time - start_time) > float(timeout):
                    self.logger.info(
                        "Drive download timeout %s seconds reached. "
                        "Stopping the download.",
                        timeout,
                    )
                    break
            return files_downloaded

        if file_dict:
            self._download_with_fileobject(file_dict)
            return [file_dict]
        return []

    @keyword(tags=["drive"])
    def update_drive_file(
        self,
        file_id: str = None,
        file_dict: dict = None,
        query: str = None,
        source: str = None,
        action: UpdateAction = UpdateAction.star,
        multiple_ok: bool = False,
    ) -> int:
        """Update file specified by id, file dictionary or query string

        Possible actions:
        - star
        - unstar
        - trash
        - untrash

        :param file_id: drive file id
        :param file_dict: file dictionary returned by `Drive Search Files`
        :param query: drive query string to find target file, needs to match 1 file
        :param source: source directory where query is executed
        :param action: update action, default star file
        :param multiple_ok: set to `True` if it is ok to perform update
         on more than 1 file
        :return: number of updated files

        Example:

        .. code-block:: robotframework

            ${folder_id}=  Get Drive Folder Id   datafolder
            ${updated}=    Update Drive File  query=name contains '.json' and '${folder_id}' in parents
            ...            action=star
            ...            multiple_ok=True
        """  # noqa: E501
        target_files = self._get_target_file(
            file_id, file_dict, query, multiple_ok, source
        )
        update_count = 0
        for tf in target_files:
            self._drive_files_update(tf, action)
            update_count += 1
        return update_count

    def _get_target_file(
        self, file_id, file_dict, query, multiple_ok, source=None, details=False
    ):
        target_files = []
        if file_id:
            if details:
                filedata = self.get_drive_file_by_id(file_id)
                target_files.append(filedata)
            else:
                target_files.append(file_id)
        elif file_dict:
            target_files.append(file_dict.get("id", None))
        else:
            files = self.search_drive_files(query, source=source, recurse=True)
            target_files = [tf if details else tf.get("id", None) for tf in files]
            if not multiple_ok and len(target_files) > 1:
                raise GoogleDriveError(
                    "expected search to match 1 file, but it matched %s files"
                    % len(files)
                )

        return target_files

    def _drive_files_update(self, file_id: str, action: UpdateAction):
        body = None
        if action == UpdateAction.trash:
            body = {"trashed": True}
        elif action == UpdateAction.untrash:
            body = {"trashed": False}
        elif action == UpdateAction.star:
            body = {"starred": True}
        elif action == UpdateAction.unstar:
            body = {"starred": False}
        else:
            # TODO: mypy should handle enum exhaustivity validation
            raise ValueError(f"Unsupported update action: {action}")
        self.logger.debug(body)
        try:
            updated_file = (
                self.service.files().update(fileId=file_id, body=body).execute()
            )
        except HttpError as err:
            raise GoogleDriveError(str(err)) from err
        return updated_file

    @keyword(tags=["drive"])
    def delete_drive_file(
        self,
        file_id: str = None,
        file_dict: dict = None,
        query: str = None,
        multiple_ok: bool = False,
        suppress_errors: bool = False,
    ) -> int:
        """Delete file specified by id, file dictionary or query string

        Note. Be extra careful when calling this keyword!

        :param file_id: drive file id
        :param file_dict: file dictionary returned by `Search Drive Files`
        :param query: drive query string to find target file, needs to match 1 file
         unless parameter `multiple_ok` is set to `True`
        :param multiple_ok: set to `True` if it is ok to perform delete
         on more than 1 file
        :param suppress_errors: on True will log warning message instead of
         raising an exception, defaults to False
        :return: how many files where deleted

        Example:

        .. code-block:: robotframework

            ${folder_id}=  Get Drive Folder Id   datafolder
            ${deleted}=    Delete Drive File  query=name contains '.json' and '${folder_id}' in parents
            ...            multiple_ok=True
        """  # noqa: E501
        target_files = self._get_target_file(file_id, file_dict, query, multiple_ok)

        delete_count = 0
        for tf in target_files:
            try:
                self.service.files().delete(fileId=tf).execute()
            except HttpError as err:
                if suppress_errors:
                    self.logger.warn(str(err))
                else:
                    raise GoogleDriveError(str(err)) from err
            delete_count += 1
        return delete_count

    @keyword(tags=["drive"])
    def get_drive_folder_id(
        self, folder: str = None, parent_folder: str = None, details: bool = False
    ) -> str:
        """Get file id for the folder

        :param folder: name of the folder to identify, by default returns drive's
         `root` folder id
        :param parent_folder: can be used to narrow search by giving parent
         folder name
        :param details: on True will return folder dictionary, on False (default)
         folder id is returned
        :return: file id of the folder or file dictionary when details = True

        Example:

        .. code-block:: robotframework

            ${root_id}=    Get Drive Folder Id   # returns Drive root folder id
            ${folder_id}=  Get Drive Folder Id  subdir
        """
        mime_folder_type = "application/vnd.google-apps.folder"
        drive_file = None
        if folder is None:
            try:
                drive_file = (
                    self.service.files().get(fileId="root", fields="id").execute()
                )
            except HttpError as err:
                raise GoogleDriveError(str(err)) from err
        else:
            query_string = f"name = '{folder}' AND mimeType = '{mime_folder_type}'"
            parameters = {"query": query_string, "recurse": True}
            if parent_folder:
                parameters["source"] = parent_folder
            folders = self.search_drive_files(**parameters)
            if len(folders) == 1:
                drive_file = folders[0]  # .get("id", None)
            else:
                self.logger.info(
                    "Found %s directories with name '%s'" % (len(folders), folder)
                )
        if drive_file:
            return drive_file if details else drive_file.get("id", None)
        return None

    @keyword(tags=["drive"])
    def move_drive_file(
        self,
        file_id: str = None,
        file_dict: dict = None,
        query: str = None,
        source: str = None,
        target: str = None,
        multiple_ok: bool = False,
    ) -> List:
        """Move file specified by id, file dictionary or query string into target folder

        :param file_id: drive file id
        :param file_dict: file dictionary returned by `Search Drive Files`
        :param query: drive query string to find target file, needs to match 1 file
        :param source: name of the folder to move file from, is by default drive's
         `root` folder id
        :param target: name of the folder to move file into, is by default drive's
         `root` folder id
        :param multiple_ok: if `True` then moving more than 1 file
        :return: list of file ids

        Example:

        .. code-block:: robotframework

            ${source_id}=  Get Drive Folder Id  sourcefolder
            ${query}=      Set Variable  name contains '.json' and '${sourceid}' in parents
            ${files}=      Move Drive File  query=${query}  folder=target_folder  multiple_ok=True
        """  # noqa: E501
        result_files = []
        if file_id or file_dict:
            target_files = self._get_target_file(file_id, file_dict, query, multiple_ok)
        else:
            target_files = self.search_drive_files(query, source=source)
        target_parent = None
        if len(target_files) == 0:
            raise GoogleDriveError("Did not find any files to move")
        if target:
            target_parent = self.get_drive_folder_id(target)
        if target_parent is None:
            raise GoogleDriveError(
                "Unable to find target folder: '%s'" % (target if target else "root")
            )
        for tf in target_files:
            file = self.service.files().get(fileId=tf["id"], fields="parents").execute()
            previous_parents = ",".join(file.get("parents"))
            try:
                result_file = (
                    self.service.files()
                    .update(
                        fileId=tf["id"],
                        addParents=target_parent,
                        removeParents=previous_parents,
                        fields="id, parents",
                    )
                    .execute()
                )
            except HttpError as err:
                raise GoogleDriveError(str(err)) from err
            result_files.append(result_file)
        return result_files

    @keyword(tags=["drive", "drive share", "v2.0.0"])
    def list_shared_drive_files(self, query: str = None, source: str = None) -> List:
        """Keyword for listing shared files in the source folder.

        Alias keyword for ``Search Drive Files`` which can be used to list
        only files which have been shared.

        :param query: drive query string to find target files
        :param source: source directory where query is executed
        :return: list of shared files

        Example:

        .. code-block:: robotframework

            ${shared}=    List Shared Drive Files    source=subfolder
            FOR    ${file}    IN    @{shared}
                Log To Console    ${file}
            END
        """
        files = self.search_drive_files(query=query, source=source)
        return [f for f in files if f["shared"]]

    @keyword(tags=["drive"])
    def search_drive_files(
        self, query: str = None, recurse: bool = False, source: str = None
    ) -> List:
        """Search Google Drive for files matching query string

        :param query: search string, defaults to None which means that all files
         and folders are returned
        :param recurse: set to `True` if search should recursive
        :param source: source directory where query is executed
        :return: list of files

        Example:

        .. code-block:: robotframework

            ${files}=  Search Drive Files   query=name contains 'hello'
            ${files}=  Search Drive Files   query=modifiedTime > '2020-06-04T12:00:00'
            ${files}=  Search Drive Files   query=mimeType contains 'image/' or mimeType contains 'video/'
            ${files}=  Search Drive Files   query=name contains '.yaml'  recurse=True
            ${files}=  Search Drive Files   query=name contains '.yaml'  source=datadirectory
        """  # noqa: E501
        page_token = None
        filelist = []

        parameters = self._set_search_parameters(query, source, recurse)

        while True:
            if page_token:
                parameters["pageToken"] = page_token
            try:
                self.logger.debug("Searching with parameters: '%s'" % parameters)
                response = self.service.files().list(**parameters).execute()
                for file_details in response.get("files", []):
                    file_dict = self._drive_file_details_into_file_dict(file_details)
                    filelist.append(file_dict)
                page_token = response.get("nextPageToken", None)
                if page_token is None:
                    break
            except HttpError as err:
                raise GoogleDriveError(str(err)) from err
        return filelist

    def _set_search_parameters(self, query, source, recurse):
        parameters = {
            "fields": "*",
        }
        if query:
            parameters["q"] = query

        folder_id = None
        if source:
            folder_id = self.get_drive_folder_id(source)
        elif not recurse:
            folder_id = "root"

        if folder_id:
            if "q" in parameters:
                parameters["q"] += f" and '{folder_id}' in parents"
            else:
                parameters["q"] = f"'{folder_id}' in parents"
        return parameters

    def _drive_file_details_into_file_dict(self, details):
        filesize = details.get("size")
        file_id = details.get("id")
        parents = details.get("parents")
        kind = details.get("kind")
        mimetype = details.get("mimeType")
        is_folder = mimetype == "application/vnd.google-apps.folder"
        folder_id = (
            file_id
            if mimetype == "application/vnd.google-apps.folder"
            else parents[0]
            if parents and len(parents) > 0
            else None
        )
        file_link = (
            None
            if mimetype == "application/vnd.google-apps.folder"
            else f"https://drive.google.com/file/d/{file_id}?usp=sharing"
        )
        folder_link = (
            f"https://drive.google.com/drive/folders/{folder_id}?usp=sharing"
            if folder_id
            else ""
        )
        file_dict = {
            "name": details.get("name"),
            "id": file_id,
            "size": int(filesize) if filesize else None,
            "kind": kind,
            "is_folder": is_folder,
            "parents": parents,
            "starred": details.get("starred"),
            "trashed": details.get("trashed"),
            "shared": details.get("shared"),
            "permissions": details.get("permissions", []),
            "mimeType": mimetype,
            "spaces": details.get("spaces", None),
            "exportLinks": details.get("exportLinks"),
            "createdTime": details.get("createdTime"),
            "modifiedTime": details.get("modifiedTime"),
            "fileLink": file_link,
            "folderLink": folder_link,
        }
        return file_dict

    @keyword(tags=["drive"])
    def create_drive_directory(
        self, folder: str = None, parent_folder: str = None
    ) -> Dict:
        """Create new directory to Google Drive

        :param folder: name for the new directory
        :param parent_folder: top level directory for new directory
        :return: dictionary containing folder ID and folder URL

        Example:

        .. code-block:: robotframework

            ${folder}=  Create Drive Directory   example-folder
            Log To Console    Google Drive folder ID: ${folder}[id]
            Log To Console    Google Drive folder URL:  ${folder}[url]
        """
        if not folder or len(folder) == 0:
            raise GoogleDriveError("Can't create Drive directory with empty name")

        folder_id = self.get_drive_folder_id(folder, parent_folder=parent_folder)
        if folder_id:
            self.logger.info(
                "Folder '%s' already exists. Not creating new one.", folder_id
            )
            return self._folder_response(folder_id)

        file_metadata = {
            "name": folder,
            "mimeType": "application/vnd.google-apps.folder",
        }

        if parent_folder:
            parent_folder_id = self.get_drive_folder_id(parent_folder)
            file_metadata["parents"] = [parent_folder_id]
        try:
            added_folder = (
                self.service.files().create(body=file_metadata, fields="id").execute()
            )
            return self._folder_response(added_folder["id"])
        except HttpError as err:
            raise GoogleDriveError(str(err)) from err

    def _folder_response(self, folder_id):
        return {
            "id": folder_id,
            "url": f"https://drive.google.com/drive/folders/{folder_id}?usp=sharing",
        }

    @keyword(tags=["drive"])
    def export_drive_file(
        self,
        file_id: str = None,
        file_dict: dict = None,
        target_file: str = None,
        mimetype: str = "application/pdf",
    ) -> str:
        """Export Google Drive file using Drive export links

        :param file_id: drive file id
        :param file_dict: file dictionary returned by `Search Drive Files`
        :param target_file: name for the exported file
        :param mimetype: export mimetype, defaults to "application/pdf"
        :return: file path to the exported file

        Example:

        .. code-block:: robotframework

            ${files}=  Drive Search Files  query=name contains 'my example worksheet'
            Export Drive File  file_dict=${files}[0]
        """
        if target_file is None or len(target_file) == 0:
            raise AttributeError("The target_file is required parameter for export")
        if file_id is None and file_dict is None:
            raise AttributeError("Either file_id or file_dict is required for export")
        target_files = self._get_target_file(file_id, file_dict, None, False)
        if len(target_files) != 1:
            raise ValueError("Did not find the Google Drive file to export")
        try:
            request = self.service.files().export(
                fileId=target_files[0], mimeType=mimetype
            )
        except HttpError as err:
            raise GoogleDriveError(str(err)) from err

        fh = BytesIO()
        downloader = MediaIoBaseDownload(fh, request)
        done = False
        while done is False:
            _, done = downloader.next_chunk()

        fh.seek(0)
        filepath = Path(target_file).absolute()
        with open(filepath, "wb") as f:
            shutil.copyfileobj(fh, f, length=131072)
        return filepath

    @keyword(tags=["drive", "drive share", "v2.0.0"])
    def add_drive_share(
        self,
        file_id: str = None,
        file_dict: dict = None,
        query: str = None,
        source: str = None,
        email: str = None,
        domain: str = None,
        role: DriveRole = DriveRole.READER,
        share_type: DriveType = DriveType.USER,
        notification: bool = False,
        notification_message: str = None,
    ) -> Dict:
        """Keyword for sharing drive file or folder.

        Parameters `file_id`, `file_dict`, `query` and `source` can be
        used to select files to which sharing is added to.

        If share is added to a folder, all files within that folder get same
        sharing permissions.

        :param file_id: drive file id
        :param file_dict: file dictionary returned by `Search Drive Files`
        :param query: drive query string to find target file, needs to match 1 file
        :param source: name of the folder to search files in, is by default drive's
         `root` folder
        :param email: user or group email address if share type
         is DriveType.USER or DriveType.GROUP
        :param domain: domain name if share type is DriveType.DOMAIN
        :param role: see ``DriveRole`` enum for possible values,
         defaults to DriveRole.READER
        :param share_type: see ``DriveType`` enum for possible values,
         defaults to DriveType.USER
        :param notification: whether to send notificatin email, defaults to False
        :param notification_message: optional message to include with the notification
        :return: share response dictionary containing 'file_id' and 'permission_id'

        Example:

        .. code-block:: robotframework

            # Add file share for a email address with email notification
            Add Drive Share
            ...    query=name = 'okta.png'
            ...    email=robocorp.tester@gmail.com
            ...    notification=True
            ...    notification_message=Hello. I have shared 'okta.png' with you for review.
            # Add file share for a domain
            Add Drive Share
            ...    query=name = 'okta.png'
            ...    domain=robocorp.com
            # Add folder share for a email address
            ${folder}=    Create Drive Directory   attachments-for-the-task
            ${share}=  Add Drive Share
            ...   file_id=${folder}[id]
            ...   email=robocorp.tester@gmail.com
            ...   role=writer
            Log To Console  Share details: ${share}[file_id], ${share}[permission_id]
        """  # noqa: E501
        target_file = self._get_target_file(
            file_id, file_dict, query, False, source=source
        )
        if not target_file:
            raise GoogleDriveError("Did not find target file")
        if domain:
            share_type = DriveType.DOMAIN
        user_permission = {
            "type": to_drive_type(share_type),
            "role": to_drive_role(role),
        }
        if share_type in [DriveType.USER, DriveType.GROUP]:
            if not email:
                raise ValueError(
                    "email parameter is required with share_type = 'user' or 'group'"
                )
            user_permission["emailAddress"] = email
        if share_type == DriveType.DOMAIN:
            if not domain:
                raise ValueError(
                    "domain parameter is required with share_type = 'domain'"
                )
            user_permission["domain"] = domain

        request_parameters = {
            "fileId": target_file[0],
            "body": user_permission,
            "fields": "id",
            "sendNotificationEmail": notification,
        }
        if notification and notification_message:
            request_parameters["emailMessage"] = notification_message

        try:
            response = self.service.permissions().create(**request_parameters).execute()
            return {"file_id": target_file[0], "permission_id": response["id"]}
        except HttpError as err:
            raise GoogleDriveError(str(err)) from err

    @keyword(tags=["drive", "drive share", "v2.0.0"])
    def remove_drive_share_by_permission_id(
        self,
        permission_id: str,
        file_id: str = None,
        file_dict: dict = None,
        query: str = None,
        source: str = None,
        suppress_errors: bool = False,
    ) -> Dict:
        """Keyword for removing share permission of file or folder
        permission id.

        Parameters `file_id`, `file_dict`, `query` and `source` can be
        used to select files from which sharing is removed.

        :param permission_id: id of the permission to remove
        :param file_id: drive file id
        :param file_dict: file dictionary returned by `Search Drive Files`
        :param query: drive query string to find target file, needs to match 1 file
        :param source: name of the folder to search files in, is by default drive's
         `root` folder
        :param suppress_errors: on True will log warning message instead of
         raising an exception, defaults to False (exception is raised)
        :return: dictionary of permission response

        Example:

        .. code-block:: robotframework

            ${share}=   Add Drive Share
            ...  query=name = 'sharable-files' and mimeType = 'application/vnd.google-apps.folder'
            ...  email=robocorp.tester@gmail.com
            #
            # actions on shared files in the folder 'shareable-files' ....
            #
            Remove Drive Share By Permission Id   ${share}[permission_id]  ${share}[file_id]
        """  # noqa: E501
        target_file = self._get_target_file(
            file_id, file_dict, query, False, source=source
        )
        if not target_file:
            raise GoogleDriveError("Did not find target file")

        self.logger.info(
            "Removing permission id '%s' for file_id '%s'"
            % (permission_id, target_file[0])
        )
        response = None
        try:
            response = (
                self.service.permissions()
                .delete(fileId=target_file[0], permissionId=permission_id)
                .execute()
            )
        except HttpError as err:
            if suppress_errors:
                self.logger.warn(str(err))
            else:
                raise GoogleDriveError(str(err)) from err
        return response

    @keyword(tags=["drive", "drive share", "v2.0.0"])
    def remove_drive_share_by_criteria(
        self,
        email: str = None,
        domain: str = None,
        permission_id: str = None,
        file_id: str = None,
        file_dict: dict = None,
        query: str = None,
        source: str = None,
        suppress_errors: bool = False,
    ) -> List:
        """Keyword for removing share from file or folder
        based on criteria.

        Parameters `file_id`, `file_dict`, `query` and `source` can be
        used to select files from which sharing is removed.

        Parameters `email`, `domain` or `permission_id` can be
        used to select which share is removed from selected files.

        :param email: email address of the permission to remove
        :param domain: domain name of the permission to remove
        :param permission_id: id of the permission to remove
        :param file_id: drive file id
        :param file_dict: file dictionary returned by `Search Drive Files`
        :param query: drive query string to find target files
        :param source: name of the folder to search files in, is by default drive's
         `root` folder
        :param suppress_errors: on True will log warning message instead of
         raising an exception, defaults to False (exception is raised)
        :return: list of dictionaries containing information of file permissions removed

        Example:

        .. code-block:: robotframework

            # Remove domain shares for files in the folder ${FOLDER_NAME}
            ${removed}=    Remove Drive Share By Criteria
            ...    domain=robocorp.com
            ...    source=${FOLDER_NAME}
            # Remove email share for a file
            ${removed}=    Remove Drive Share By Criteria
            ...    query=name = 'okta.png'
            ...    email=robocorp.tester@gmail.com
        """
        if not email and not domain and permission_id:
            raise AttributeError(
                "At least one of the 'email', 'domain' or 'permission_id' "
                "is required to remove drive share"
            )
        target_files = self._get_target_file(
            file_id, file_dict, query, multiple_ok=True, source=source, details=True
        )
        if not target_files or len(target_files) == 0:
            raise GoogleDriveError("Did not find target files")
        permissions_removed = []
        for tf in target_files:
            file_permissions_removed = []
            if "permissions" in tf and tf["permissions"]:
                self.logger.info(
                    "Removing shares from file '%s' id '%s'" % (tf["name"], tf["id"])
                )
                for p in tf["permissions"]:
                    if self._is_permission_removable(p, email, domain, permission_id):
                        self._remove_file_permission(
                            tf, p, file_permissions_removed, suppress_errors
                        )
            if file_permissions_removed:
                permissions_removed.append(
                    {
                        "file_id": tf["id"],
                        "file_name": tf["name"],
                        "permissions_removed": file_permissions_removed,
                    }
                )
        return permissions_removed

    def _is_permission_removable(self, permission, email, domain, permission_id):
        return (
            (
                email
                and "emailAddress" in permission.keys()
                and permission["emailAddress"] == email
            )
            or (
                domain
                and "domain" in permission.keys()
                and permission["domain"] == domain
            )
            or (permission_id and permission["id"] == permission_id)
        )

    def _remove_file_permission(
        self, drive_file, permission, permissions_removed, suppress_errors
    ):
        try:
            self.service.permissions().delete(
                fileId=drive_file["id"], permissionId=permission["id"]
            ).execute()
            permissions_removed.append(permission)
        except HttpError as err:
            if suppress_errors:
                self.logger.warn(str(err))
            else:
                raise GoogleDriveError(str(err)) from err

    @keyword(tags=["drive", "drive share", "v2.0.0"])
    def remove_all_drive_shares(
        self,
        file_id: str = None,
        file_dict: dict = None,
        query: str = None,
        suppress_errors: bool = False,
    ) -> List:
        """Keyword for removing all shares from selected files (only owner
        permission is retained).

        :param file_id: drive file id
        :param file_dict: file dictionary returned by `Search Drive Files`
        :param query: drive query string to find target files
        :param suppress_errors: on True will log warning message instead of
         raising an exception, defaults to False (exception is raised)
        :return: list of dictionaries containing information of file permissions removed

        Example:

        .. code-block:: robotframework

            ${removed}=  Remove All Drive Shares    file_id=${FOLDER_ID}
        """
        target_files = self._get_target_file(
            file_id, file_dict, query, multiple_ok=True, details=True
        )
        permissions_removed = []
        for tf in target_files:
            if "permissions" in tf and tf["permissions"]:
                self.logger.info(
                    "Removing shares from file '%s' id '%s'" % (tf["name"], tf["id"])
                )
                for p in tf["permissions"]:
                    if p["role"] != "owner":
                        self.remove_drive_share_by_permission_id(
                            file_id=tf["id"],
                            permission_id=p["id"],
                            suppress_errors=suppress_errors,
                        )
                        permissions_removed.append(p)
        return permissions_removed

    @keyword(tags=["drive", "v2.0.0"])
    def get_drive_file_by_id(self, file_id: str, suppress_errors: bool = False) -> Dict:
        """Get file dictionary by its file id.

        :param file_id: id of the file in the Google Drive
        :param suppress_errors: on True will log warning message instead of
         raising an exception, defaults to False (exception is raised)
        :return: dictionary containing file information

        Example:

        .. code-block:: robotframework

            ${file_dict}=  Get Drive File By ID    file_id=${FILE_ID}
        """
        response = None
        try:
            raw_response = (
                self.service.files().get(fileId=file_id, fields="*").execute()
            )
            response = self._drive_file_details_into_file_dict(raw_response)
        except HttpError as err:
            if suppress_errors:
                self.logger.warn(str(err))
            else:
                raise GoogleDriveError(str(err)) from err
        return response



from typing import Any, Dict, List, Optional

from google.cloud import storage

from . import (
    LibraryContext,
    keyword,
)


class StorageKeywords(LibraryContext):
    """Class for Google Cloud Storage API
     and Google Cloud Storage JSON API

    Link to `Google Storage PyPI`_ page.

    .. _Google Storage PyPI: https://pypi.org/project/google-cloud-storage/
    """

    def __init__(self, ctx):
        super().__init__(ctx)
        self.service = None

    @keyword(tags=["init", "storage"])
    def init_storage(
        self,
        service_account: str = None,
        use_robocorp_vault: Optional[bool] = None,
        token_file: str = None,
    ) -> None:
        """Initialize Google Cloud Storage client

        :param service_account: file path to service account file
        :param use_robocorp_vault: use credentials in `Robocorp Vault`
        :param token_file: file path to token file
        """
        self.service = self.init_service_with_object(
            storage.Client, service_account, use_robocorp_vault, token_file
        )

    @keyword(tags=["storage"])
    def create_storage_bucket(self, bucket_name: str) -> Dict:
        """Create Google Cloud Storage bucket

        :param bucket_name: name as string
        :return: bucket

        **Examples**

        **Robot Framework**

        .. code-block:: robotframework

            ${result}=   Create Storage Bucket   visionfolder
        """
        bucket = self.service.create_bucket(bucket_name)
        return bucket

    @keyword(tags=["storage"])
    def delete_storage_bucket(self, bucket_name: str):
        """Delete Google Cloud Storage bucket

        Bucket needs to be empty before it can be deleted.

        :param bucket_name: name as string

        **Examples**

        **Robot Framework**

        .. code-block:: robotframework

            ${result}=   Delete Storage Bucket   visionfolder
        """
        bucket = self.get_storage_bucket(bucket_name)
        try:
            bucket.delete()
        except Exception as e:
            raise ValueError("The bucket you tried to delete was not empty") from e

    @keyword(tags=["storage"])
    def get_storage_bucket(self, bucket_name: str) -> Dict:
        """Get Google Cloud Storage bucket

        :param bucket_name: name as string
        :return: bucket

        **Examples**

        **Robot Framework**

        .. code-block:: robotframework

            ${result}=   Get Bucket   visionfolder
        """
        bucket = self.service.get_bucket(bucket_name)
        return bucket

    @keyword(tags=["storage"])
    def list_storage_buckets(self) -> List:
        """List Google Cloud Storage buckets

        :return: list of buckets

        **Examples**

        **Robot Framework**

        .. code-block:: robotframework

            ${buckets}=   List Storage Buckets
            FOR  ${bucket}  IN   @{buckets}
                Log  ${bucket}
            END
        """
        return list(self.service.list_buckets())

    @keyword(tags=["storage"])
    def delete_storage_files(self, bucket_name: str, files: Any) -> List:
        """Delete files in the bucket

        Files need to be object name in the bucket.

        :param bucket_name: name as string
        :param files: single file, list of files or comma separated list of files
        :return: list of files which could not be deleted

         **Examples**

        **Robot Framework**

        .. code-block:: robotframework

            ${result}=   Delete Storage Files   ${BUCKET_NAME}   file1,file2
        """
        if not isinstance(files, list):
            files = files.split(",")
        bucket = self.get_storage_bucket(bucket_name)
        notfound = []
        for filename in files:
            filename = filename.strip()
            blob = bucket.get_blob(filename)
            if blob:
                blob.delete()
            else:
                notfound.append(filename)
        return notfound if len(notfound) > 0 else True

    @keyword(tags=["storage"])
    def list_storage_files(self, bucket_name: str) -> List:
        """List files in the bucket

        :param bucket_name: name as string
        :return: list of object names in the bucket

        **Examples**

        **Robot Framework**

        .. code-block:: robotframework

            ${files}=   List Storage Files  ${BUCKET_NAME}
            FOR  ${bucket}  IN   @{files}
                Log  ${file}
            END
        """
        bucket = self.get_storage_bucket(bucket_name)
        all_blobs = bucket.list_blobs()
        sorted_blobs = sorted(blob.name for blob in all_blobs)
        return [
            {"name": name, "uri": f"gs://{bucket_name}/{name}"} for name in sorted_blobs
        ]

    @keyword(tags=["storage"])
    def upload_storage_file(
        self, bucket_name: str, filename: str, target_name: str
    ) -> None:
        """Upload a file into a bucket

        :param bucket_name: name as string
        :param filename: filepath to upload file
        :param target_name: target object name

        **Examples**

        **Robot Framework**

        .. code-block:: robotframework

            Upload Storage File  ${BUCKET_NAME}
            ...   ${CURDIR}${/}test.txt    test.txt
        """
        bucket = self.get_storage_bucket(bucket_name)
        blob = bucket.blob(target_name)
        with open(filename, "rb") as f:
            blob.upload_from_file(f)

    @keyword(tags=["storage"])
    def upload_storage_files(self, bucket_name: str, files: dict) -> None:
        """Upload files into a bucket

        Example `files`:
        files = {"mytestimg": "image1.png", "mydoc": "google.pdf"}

        :param bucket_name: name as string
        :param files: dictionary of object names and filepaths

        **Examples**

        **Robot Framework**

        .. code-block:: robotframework

            ${files}=   Create Dictionary
            ...   test1.txt   ${CURDIR}${/}test1.txt
            ...   test2.txt   ${CURDIR}${/}test2.txt
            Upload Storage Files   ${BUCKET_NAME}   ${files}
        """
        if not isinstance(files, dict):
            raise ValueError("files needs to be an dictionary")
        bucket = self.get_storage_bucket(bucket_name)
        for target_name, filename in files.items():
            blob = bucket.blob(target_name)
            blob.upload_from_filename(filename)

    @keyword(tags=["storage"])
    def download_storage_files(self, bucket_name: str, files: Any) -> List:
        """Download files from a bucket

        Example `files`:
        files = {"mytestimg": "image1.png", "mydoc": "google.pdf"}

        :param bucket_name: name as string
        :param files: list of object names or dictionary of
            object names and target files
        :return: list of files which could not be downloaded

        **Examples**

        **Robot Framework**

        .. code-block:: robotframework

            ${result}=  Download Storage Files  ${BUCKET_NAME}   test1.txt,test2.txt
        """
        if isinstance(files, str):
            files = files.split(",")
        bucket = self.get_storage_bucket(bucket_name)
        notfound = []
        if isinstance(files, dict):
            for object_name, filename in files.items():
                blob = bucket.get_blob(object_name)
                if blob:
                    with open(filename, "wb") as f:
                        blob.download_to_file(f)
                        self.logger.info(
                            "Downloaded object %s from Google to filepath %s",
                            object_name,
                            filename,
                        )
                else:
                    notfound.append(object_name)
        else:
            for filename in files:
                filename = filename.strip()
                blob = bucket.get_blob(filename)
                if blob:
                    with open(filename, "wb") as f:
                        blob.download_to_file(f)
                        self.logger.info(
                            "Downloaded object %s from Google to filepath %s",
                            filename,
                            filename,
                        )
                else:
                    notfound.append(filename)
        return notfound



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['RPA', 'RPA.Cloud.Google', 'RPA.Cloud.Google.keywords', 'RPA.scripts']

package_data = \
{'': ['*']}

install_requires = \
['google-api-python-client>=2.58.0,<3.0.0',
 'google-auth-httplib2>=0.1.0,<0.2.0',
 'google-auth-oauthlib>=0.5.2,<0.6.0',
 'google-cloud-documentai>=2.0.1,<3.0.0',
 'google-cloud-language>=2.5.2,<3.0.0',
 'google-cloud-speech>=2.15.1,<3.0.0',
 'google-cloud-storage>=2.5.0,<3.0.0',
 'google-cloud-texttospeech>=2.12.1,<3.0.0',
 'google-cloud-translate>=3.8.1,<4.0.0',
 'google-cloud-videointelligence>=2.8.1,<3.0.0',
 'google-cloud-vision>=3.1.1,<4.0.0',
 'grpcio>=1.48.1,<2.0.0',
 'robotframework-pythonlibcore>=3.0.0,<4.0.0',
 'robotframework>=4.0.0,!=4.0.1,<6.0.0',
 'rpaframework-core>=10.0.0,<11.0.0']

entry_points = \
{'console_scripts': ['rpa-google-oauth = RPA.scripts.google_authenticate:main']}

setup_kwargs = {
    'name': 'rpaframework-google',
    'version': '6.1.0',
    'description': 'Google library for RPA Framework',
    'long_description': 'rpaframework-google\n===================\n\nThis library enables Google Cloud services for `RPA Framework`_\nlibraries, such as Google Sheets and Google Vision.\n\n.. _RPA Framework: https://rpaframework.org\n',
    'author': 'RPA Framework',
    'author_email': 'rpafw@robocorp.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://rpaframework.org/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import argparse
import base64
import os
import pickle
import sys

from google_auth_oauthlib.flow import InstalledAppFlow


DESCRIPTION = """
Command-line utility for converting Google OAuth
authorization credentials to an access token, which can be kept
in a secret storage such as Robocorp Vault. This allows calling Google
APIs without user interaction.

example usage:
> rpa-google-oauth --credentials drive_credentials.json --service drive
"""

CREDENTIALS = """
Copy these credentials into Robocorp Vault:

-------------------------------------------
{}
-------------------------------------------
"""

SERVICE_SCOPES = {
    "drive": ["drive.appdata", "drive.file", "drive.install", "drive"],
    "natural-language": ["cloud-language", "cloud-platform"],
    "sheets": ["spreadsheets"],
    "apps-script": ["script.projects"],
    "vision": ["cloud-vision", "cloud-platform"],
    "gmail": ["gmail.send", "gmail.compose", "gmail.modify", "gmail.labels"],
}


def create_parser():
    parser = argparse.ArgumentParser(
        description=DESCRIPTION.strip(), formatter_class=argparse.RawTextHelpFormatter
    )
    parser.add_argument(
        "--credentials",
        dest="credentials_file",
        default="credentials.json",
        help="project credentials file (default: %(default)s)",
    )
    parser.add_argument(
        "--scopes",
        dest="scopes",
        help="authentication scopes as comma separated list",
    )
    parser.add_argument(
        "--service",
        dest="service",
        help="set authentication scopes for the given service",
    )
    parser.add_argument(
        "--console",
        dest="console_flow",
        action="store_true",
        help="run console based auth flow (default: %(default)s)",
    )
    return parser


def main():
    parser = create_parser()
    args = parser.parse_args()

    def error(msg):
        parser.print_help()
        print(f"\nERROR: {msg}\n")
        sys.exit(1)

    # if Google.GOOGLECLOUD_IMPORT_ERROR:
    #     error(
    #         "Please install the 'google' optional extra to use this script:\n"
    #         "> pip install rpaframework[google]"
    #     )

    if (
        not os.path.exists(args.credentials_file)
        or os.stat(args.credentials_file).st_size == 0
    ):
        error(f"Credentials file '{args.credentials_file}' does not exist or is empty")

    scopes = []

    if args.service:  # and args.service in SERVICE_SCOPES:
        services = args.service.split(",")
        for service in services:
            if service in SERVICE_SCOPES.keys():
                scopes += SERVICE_SCOPES[service]

    if args.scopes:
        scopes.extend(args.scopes.split(","))

    if not scopes:
        error("No authentication scopes have been defined")

    print(scopes)
    uris = [f"https://www.googleapis.com/auth/{scope}" for scope in scopes]
    print("Google OAuth Flow for scopes: {}".format(", ".join(scopes)))

    flow = InstalledAppFlow.from_client_secrets_file(args.credentials_file, uris)
    if args.console_flow:
        credentials = flow.run_console()
    else:
        credentials = flow.run_local_server()

    serialized = base64.b64encode(pickle.dumps(credentials)).decode("utf-8")
    print(CREDENTIALS.format(serialized))


if __name__ == "__main__":
    main()
